import re
from types import SimpleNamespace
from typing import Literal, Optional, TypedDict
from ..constant import Gender
from ..util import CHECK_DIGIT, weighted_modulus_digit, validate_regexp


class ParseResult(TypedDict):
    location: Literal['A', 'B', 'C', 'D', 'E', 'F', 'G',
                      'H', 'I', 'J', 'K', 'L', 'M', 'N',
                      'O', 'P', 'Q', 'R', 'S', 'T', 'U',
                      'V', 'W', 'X', 'Y', 'Z']
    """location code"""
    gender: Gender
    """gender: male or female"""
    sn: str
    """serial number"""
    checksum: CHECK_DIGIT


class NationalID:
    """
    TWN National ID number format
    https://en.wikipedia.org/wiki/National_identification_number#Taiwan
    https://zh.wikipedia.org/wiki/%E4%B8%AD%E8%8F%AF%E6%B0%91%E5%9C%8B%E5%9C%8B%E6%B0%91%E8%BA%AB%E5%88%86%E8%AD%89
    python version of http://www2.lssh.tp.edu.tw/~hlf/class-1/lang-c/id/index.htm
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'TW',
        'min_length': 10,
        'max_length': 10,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<location>[A-Z])'
                             r'(?P<gender>[12])'
                             r'(?P<sn>\d{7})'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['National ID Number',
                  '國民身分證統一編號',
                  '身分證字號'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Taiwan',
                  'https://zh.wikipedia.org/wiki/'
                  '%E4%B8%AD%E8%8F%AF%E6%B0%91%E5%9C%8B%E5%9C%8B%E6%B0%91%E8%BA%AB%E5%88%86%E8%AD%89'],
        'deprecated': False
    })

    LOCATION_NUM = [[1, 0], [1, 1], [1, 2], [1, 3], [1, 4], [1, 5], [1, 6],
                    [1, 7], [3, 4], [1, 8], [1, 9], [2, 0], [2, 1], [2, 2],
                    [3, 5], [2, 3], [2, 4], [2, 5], [2, 6], [2, 7], [2, 8],
                    [2, 9], [3, 2], [3, 0], [3, 1], [3, 3]]
    """A-Z to numbers by index"""

    MAGIC_MULTIPLIER = [1, 9, 8, 7, 6, 5, 4, 3, 2, 1]
    """multiplier for checksum"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the TWN id number
        """
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        return NationalID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the value"""
        match_obj = NationalID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        checksum = NationalID.checksum(id_number)
        if str(checksum) != id_number[-1]:
            return None
        location = match_obj.group('location')
        gender = match_obj.group('gender')
        sn = match_obj.group('sn')
        return {
            'location': location,
            'gender': Gender.MALE if gender == '1' else Gender.FEMALE,
            'sn': sn,
            'checksum': int(match_obj.group('checksum'))
        }

    @staticmethod
    def checksum(id_number) -> Optional[CHECK_DIGIT]:
        """
        algorithm:
        https://zh.wikipedia.org/wiki/%E4%B8%AD%E8%8F%AF%E6%B0%91%E5%9C%8B%E5%9C%8B%E6%B0%91%E8%BA%AB%E5%88%86%E8%AD%89#%E6%9C%89%E6%95%88%E7%A2%BC
        """
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        # it uses modulus 10 algorithm with magic numbers
        location = id_number[0]
        numbers = NationalID.LOCATION_NUM[ord(location) - 65] + [int(char) for char in id_number[1:]]
        modulus = weighted_modulus_digit(numbers[:-1], NationalID.MAGIC_MULTIPLIER, 10)
        return modulus



import re
from types import SimpleNamespace
from ..util import validate_regexp, weighted_modulus_digit


class NationalID:
    """
    Turkey National ID number
    https://en.wikipedia.org/wiki/National_identification_number#Turkey
    https://stackoverflow.com/questions/53610208/turkish-identity-number-verification
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'TR',
        # length without insignificant chars
        'min_length': 11,
        'max_length': 11,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r'^[1-9]\d{10}$'),
        'alias_of': None,
        'names': ['National ID Number',
                  'Türkiye Cumhuriyeti Kimlik Numarası',
                  'T.C. Kimlik No.'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Turkey'],
        'deprecated': False
    })

    MULTIPLIERS = [7, -1, 7, -1, 7, -1, 7, -1, 7]
    """multiplier for the checksum"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate TUR national id number
        """
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        return NationalID.checksum(id_number) == id_number[-2:]

    @staticmethod
    def checksum(id_number: str) -> str:
        """
        Calculate the checksum e.g. digit 10 and digit 11
        """
        numbers_list = [int(i) for i in id_number]
        digit_ten = weighted_modulus_digit(numbers_list[:-2], NationalID.MULTIPLIERS, 10, True)
        digit_eleven = sum(numbers_list[:-1]) % 10
        return f'{digit_ten}{digit_eleven}'



import re
from types import SimpleNamespace
from typing import Literal, Optional
from ..util import weighted_modulus_digit, modulus_overflow_mod10, validate_regexp
from .resident_registration import ResidentRegistration, ParseResult


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r'-', '', id_number)


class OldIDParseResult(ParseResult):
    """Old format contains more information"""
    location: str
    """registration location"""
    checksum: Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    """check digits"""


class OldResidentRegistration(ResidentRegistration):
    """
    KOR National ID number format. The ARC is the same as NationalID.
    # https://en.wikipedia.org/wiki/Resident_registration_number
    # https://centers.ibs.re.kr/html/living_en/overview/arc.html
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'KR',
        'min_length': 13,
        'max_length': 13,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<yy>\d{2})(?P<mm>\d{2})(?P<dd>\d{2})-'
                             r'(?P<gender>\d)'
                             r'(?P<location>\d{4})'
                             r'(?P<sn>\d)'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['Resident Registration Number',
                  '주민등록번호',
                  'RRN',
                  '住民登錄番號',
                  'Jumin Deungnok Beonho'],
        'links': ['https://en.wikipedia.org/wiki/Resident_registration_number',
                  'https://centers.ibs.re.kr/html/living_en/overview/arc.html'],
        'deprecated': True
    })

    MAGIC_MULTIPLIER = [2, 3, 4, 5, 6, 7, 8, 9, 2, 3, 4, 5]
    """multiplier for the checksum"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the old KOR id number
        """
        if not validate_regexp(id_number, OldResidentRegistration.METADATA.regexp):
            return False
        return OldResidentRegistration.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[OldIDParseResult]:
        """prase the result"""
        match_obj = re.match(OldResidentRegistration.METADATA.regexp, id_number)
        new_result = ResidentRegistration.parse(id_number)
        if not new_result:
            return None
        if not OldResidentRegistration.checksum(id_number):
            return None
        return {
            **new_result,
            'sn': match_obj.group('sn'),
            'location': match_obj.group('location'),
            'checksum': int(match_obj.group('checksum'))
        }

    @staticmethod
    def checksum(id_number) -> bool:
        """multiply the magic number and find the modulus"""
        if not validate_regexp(id_number, OldResidentRegistration.METADATA.regexp):
            return False
        normalized = normalize(id_number)
        # it uses modulus 11 algorithm with magic numbers
        numbers = [int(char) for char in normalized]
        modulus = modulus_overflow_mod10(
            weighted_modulus_digit(numbers[:-1], OldResidentRegistration.MAGIC_MULTIPLIER, 11))
        return modulus == numbers[-1]



import re
from datetime import date, timedelta
from types import SimpleNamespace
from typing import List, Optional, TypedDict
from ..constant import Gender
from ..util import validate_regexp


class TaxpayerIDParseResult(TypedDict):
    """parse result of the taxpayer id"""
    yyyymmdd: date
    gender: Gender
    checksum: int


class TaxpayerIDNumber:
    """
    Ukraine Taxpayer ID number format
    https://en.wikipedia.org/wiki/National_identification_number#Ukraine
    This is a python version of https://github.com/therezor/ua-tax-number/blob/main/src/Decoder.php
    The alias: ['RNTRC', 'РНОКПП', 'taxpayer registration number']
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'UA',
        # length without insignificant chars
        'min_length': 10,
        'max_length': 10,
        # has parse function
        'parsable': True,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^\d{10}$'),
        'alias_of': None,
        'names': ['Taxpayer ID Number',
                  'RNTRC',
                  'РНОКПП',
                  'taxpayer registration number'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Ukraine'],
        'deprecated': False
    })

    MAGIC_MULTIPLIER: List[int] = [-1, 5, 7, 9, 4, 6, 10, 5, 7]
    """multiplier for the checksum"""
    BIRTHDAY_BASE: date = date(1900, 1, 1)

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the id number
        """
        if not validate_regexp(id_number, TaxpayerIDNumber.METADATA.regexp):
            return False

        return TaxpayerIDNumber.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[TaxpayerIDParseResult]:
        """parse the result"""
        if TaxpayerIDNumber.checksum(id_number) != int(id_number[9]):
            return None
        # according to the PHP implementation, we need to minus 1, maybe the tail and head values included.
        days = int(id_number[:5]) - 1
        dob = TaxpayerIDNumber.BIRTHDAY_BASE + timedelta(days=days)
        return {
            'yyyymmdd': dob,
            'gender': Gender.MALE if int(id_number[8]) % 2 == 1 else Gender.FEMALE,
            'checksum': int(id_number[9])
        }

    @staticmethod
    def checksum(id_number: str) -> Optional[int]:
        """algorithm: https://github.com/therezor/ua-tax-number/blob/main/src/Decoder.php"""
        if not validate_regexp(id_number, TaxpayerIDNumber.METADATA.regexp):
            return None
        number_list = [int(char) for char in list(id_number)]
        source_list = number_list[:9]
        total = sum([value * TaxpayerIDNumber.MAGIC_MULTIPLIER[index] for (index, value) in enumerate(source_list)])
        # calculate the modulus, if the value is 10, use the 0. Will it collide?
        return total % 11 % 10



import re
from datetime import date
from types import SimpleNamespace
from typing import Optional, TypedDict, Tuple
from ..util import CHECK_DIGIT, weighted_modulus_digit, validate_regexp
from ..constant import Citizenship, Gender


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r'[ -]', '', id_number)


class ParseResult(TypedDict):
    """parse result for the national id"""
    yyyymmdd: date
    """birthday"""
    gender: Gender
    """gender, possible value: male, female"""
    citizenship: Citizenship
    """possible value: citizen meaning born from hungary, foreign: meaning naturalized citizen or resident"""
    sn: str
    """serial number"""
    checksum: CHECK_DIGIT
    """check digits"""


class PersonalID:
    """
    Hungary Personal ID number format
    https://en.wikipedia.org/wiki/National_identification_number#Hungary
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'HU',
        'min_length': 11,
        'max_length': 11,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<gender>\d)[ -]?'
                             r'(?P<yy>\d{2})(?P<mm>\d{2})(?P<dd>\d{2})[ -]?'
                             r'(?P<sn>\d{3})'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['Personal ID Number'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Hungary'],
        'deprecated': False
    })

    MAGIC_MULTIPLIER = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    """multiplier for the checksum"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the personal id number
        """
        if not validate_regexp(id_number, PersonalID.METADATA.regexp):
            return False
        return PersonalID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the result"""
        match_obj = PersonalID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        checksum = PersonalID.checksum(id_number)
        if not checksum:
            return None
        gender_citizenship_year_base = PersonalID.get_gender_citizenship_year_base(int(match_obj.group('gender')))
        if not gender_citizenship_year_base:
            return None
        gender, citizenship, year_base = gender_citizenship_year_base
        yy = int(match_obj.group('yy'))
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        sn = match_obj.group('sn')
        try:
            return {
                'yyyymmdd': date(yy + year_base, mm, dd),
                'gender': gender,
                'citizenship': citizenship,
                'sn': sn,
                'checksum': int(match_obj.group('checksum'))
            }
        except ValueError:
            return None

    @staticmethod
    def checksum(id_number) -> bool:
        """
        It's hard to find it. The algorithm is the 11-modulus on a weighted sum.
        algorithm: https://github.com/loonkwil/hungarian-validator-bundle/blob/master/Validator/PersonalIdValidator.php
        """
        if not validate_regexp(id_number, PersonalID.METADATA.regexp):
            return False
        # it uses modulus 11 algorithm with magic numbers
        numbers = [int(char) for char in normalize(id_number)]
        modulus = weighted_modulus_digit(numbers[:-1], PersonalID.MAGIC_MULTIPLIER, 11, True)
        # According to an official doc in hungary language, gov will use another random number to
        # skip the modulus 10.
        return modulus == numbers[-1] if modulus < 10 else False

    @staticmethod
    def get_gender_citizenship_year_base(gender_citizenship: int) -> Optional[Tuple[Gender, Citizenship, int]]:
        gender = Gender.MALE if gender_citizenship % 2 == 1 else Gender.FEMALE
        citizenship = Citizenship.CITIZEN if gender_citizenship < 5 else Citizenship.FOREIGN
        if 0 < gender_citizenship <= 2:
            year_base = 1900
        elif 2 < gender_citizenship <= 4:
            year_base = 2000
        elif 4 < gender_citizenship <= 6:
            year_base = 1900
        elif 6 < gender_citizenship <= 8:
            year_base = 1800
        else:
            return None
        return gender, citizenship, year_base



import re
from types import SimpleNamespace
from typing import Optional
from ..util import CHECK_DIGIT, validate_regexp, weighted_modulus_digit


def normalize(id_number: str) -> str:
    """strip out useless characters/whitespaces"""
    return id_number.replace('-', '')


class NationalID:
    """
    Iran national id number, (کارت ملی/kart-e-meli)
    https://en.wikipedia.org/wiki/National_identification_number#Iran,_Islamic_Republic_of
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'IR',
        'min_length': 10,
        'max_length': 10,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r'^\d{3}-?\d{6}-?\d$'),
        'alias_of': None,
        'names': ['National ID Number',
                  'kart-e-meli',
                  'کارت ملی'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Iran,_Islamic_Republic_of'],
        'deprecated': False
    })

    MULTIPLIER = [10, 9, 8, 7, 6, 5, 4, 3, 2]

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate
        """
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        return NationalID.checksum(id_number) == int(id_number[-1])

    @staticmethod
    def checksum(id_number: str) -> Optional[CHECK_DIGIT]:
        """algorithm: https://github.com/mohammadv184/idvalidator/blob/main/validate/nationalid/nationalid.go"""
        normalized = normalize(id_number)
        numbers = [int(i) for i in normalized]
        modulus = weighted_modulus_digit(numbers[:-1], NationalID.MULTIPLIER, 11, True)
        return modulus if modulus < 2 else 11 - modulus



import re
from datetime import date
from types import SimpleNamespace
from typing import Literal, Optional, TypedDict
from ..constant import Gender
from ..util import validate_regexp


def normalize(id_number: str) -> str:
    """strip out useless characters/whitespaces"""
    return id_number.upper() if id_number else None


class ParseResult(TypedDict):
    """The parse result of Resident ID"""
    address_code: str
    """registration address code"""
    yyyymmdd: date
    """birthday"""
    sn: str
    """serial number"""
    gender: Gender
    """gender, possible value: male, female"""
    checksum: Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 'X']
    """checksum digit, numbers or 'X'"""


class ResidentID:
    """
    China Resident ID number format
    https://en.wikipedia.org/wiki/Resident_Identity_Card
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'CN',
        'min_length': 18,
        'max_length': 18,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<address_code>\d{6})'
                             r'(?P<yyyy>\d{4})'
                             r'(?P<mm>0[1-9]|1[12])'
                             r'(?P<dd>0[1-9]|[12][0-9]|3[01])'
                             r'(?P<sn>\d{3})'
                             r'(?P<checksum>(\d|X))$'),
        'alias_of': None,
        'names': ['Resident Identity Number',
                  '居民身份证',
                  'Jūmín Shēnfènzhèng'],
        'links': ['https://en.wikipedia.org/wiki/Resident_Identity_Card',
                  'https://en.wikipedia.org/wiki/National_identification_number#China'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the CHN id number
        """
        if not id_number:
            return False

        if not isinstance(id_number, str):
            id_number = repr(id_number)
        return ResidentID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the data"""
        match_obj = ResidentID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        address_code = match_obj.group('address_code')
        sn = match_obj.group('sn')
        checksum_str = match_obj.group('checksum')
        checksum = ResidentID.checksum(id_number)
        if checksum is None or str(checksum) != checksum_str:
            return None
        try:
            return {
                'address_code': address_code,
                'yyyymmdd': date(int(match_obj.group('yyyy')), int(match_obj.group('mm')), int(match_obj.group('dd'))),
                'sn': sn,
                'gender': Gender.FEMALE if int(sn) % 2 == 0 else Gender.MALE,
                'checksum': checksum
            }
        except ValueError:
            return None

    @staticmethod
    def checksum(id_number) -> Optional[Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 'X']]:
        """algorithm: https://en.wikipedia.org/wiki/Resident_Identity_Card#Identity_card_number"""
        if not validate_regexp(id_number, ResidentID.METADATA.regexp):
            return None
        normalized = normalize(id_number)
        source_list = [int(char) for char in normalized[:-1]]
        # The magic number is calculated from 2^(18 - i) % 11 (the first number is 1).
        total = sum([value * (pow(2, 17 - index) % 11) for (index, value) in enumerate(source_list)])
        checksum = (12 - total % 11) % 11
        return 'X' if checksum == 10 else checksum



import re
from types import SimpleNamespace
from typing import List
from ..util import validate_regexp


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r'-', '', id_number)


class InlandRevenueDepartmentNumber:
    """
    New Zealand inland revenue department(IRD) number format
    https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/New%20Zealand-TIN.pdf
    This is a python version of this one: https://github.com/jarden-digital/nz-ird-validator
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'NZ',
        # length without insignificant chars
        'min_length': 8,
        'max_length': 9,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^('
                             r'\d{9}|\d{3}-\d{3}-\d{3}|'
                             r'\d{8}|\d{2}-\d{3}-\d{3}'
                             r')$'),
        'alias_of': None,
        'names': ['Inland Revenue Department Number',
                  'IRD'],
        'links': ['https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/'
                  'tax-identification-numbers/New%20Zealand-TIN.pdf'],
        'deprecated': False
    })

    PHASE1_MULTIPLIER = [3, 2, 7, 6, 5, 4, 3, 2]
    PHASE2_MULTIPLIER = [7, 4, 3, 2, 5, 2, 7, 6]

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the NZL IRD number
        """
        return InlandRevenueDepartmentNumber.checksum(id_number)

    @staticmethod
    def checksum(id_number: str) -> bool:
        """algorithm: https://github.com/jarden-digital/nz-ird-validator"""
        if not validate_regexp(id_number, InlandRevenueDepartmentNumber.METADATA.regexp):
            return False
        normalized = normalize(id_number)
        if len(normalized) == 8:
            # pre-pad a 0 if it is the short one
            normalized = '0' + normalized
        number_list = [int(char) for char in list(normalized)]
        # split to source list and check digit
        source_list = number_list[:8]
        check_digit = number_list[8]
        # phase 1
        calculated = InlandRevenueDepartmentNumber.calc_checkdigit(source_list,
                                                                   InlandRevenueDepartmentNumber.PHASE1_MULTIPLIER)
        if calculated != 10:
            return calculated == check_digit
        # phase 2
        calculated2 = InlandRevenueDepartmentNumber.calc_checkdigit(source_list,
                                                                    InlandRevenueDepartmentNumber.PHASE2_MULTIPLIER)
        return (calculated2 == check_digit) if calculated2 < 10 else False

    @staticmethod
    def calc_checkdigit(source_list: List[int], magic_numbers: List[int]) -> int:
        modulus = sum([value * magic_numbers[index] for (index, value) in enumerate(source_list)]) % 11
        return 0 if modulus == 0 else (11 - modulus)



import re
from types import SimpleNamespace
from typing import Optional
from ..util import CHECK_DIGIT, validate_regexp
from .util import normalize


class MedicareNumber:
    """
    Australia medicare number format
    https://techdocs.broadcom.com/us/en/symantec-security-software/information-security/data-loss-prevention/15-8/about-data-loss-prevention-policies-v27576413-d327e9/library-of-system-data-identifiers-v95989112-d327e56315/australian-medicare-number-v115447646-d327e57399.html
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'AU',
        # length without insignificant chars
        'min_length': 9,
        'max_length': 11,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^('
                             r'[2-6]\d{10}|[2-6]\d{3} \d{5} \d|[2-6]\d{3}-\d{5}-\d|'
                             r'[2-6]\d{9}|[2-6]\d{9}([-/]\d)?|'
                             r'[2-6]\d{3} \d{5} \d([-/]\d)?|[2-6]\d{3}-\d{5}-\d([-/]\d)?|'
                             r'[2-6]\d{3} \d{5} \d \d|[2-6]\d{3}-\d{5}-\d-\d'
                             r')$'),
        'alias_of': None,
        'names': ['Medicare Number', 'Medicare No'],
        'links': [
            'https://techdocs.broadcom.com/us/en/symantec-security-software/'
            'information-security/data-loss-prevention/15-8/'
            'about-data-loss-prevention-policies-v27576413-d327e9'
            '/library-of-system-data-identifiers-v95989112-d327e56315/'
            'australian-medicare-number-v115447646-d327e57399.html'],
        'deprecated': False
    })

    MAGIC_MULTIPLIER = [1, 3, 7, 9, 1, 3, 7, 9]
    """magic multiplier for checksum"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the medicare number
        """
        if not validate_regexp(id_number, MedicareNumber.METADATA.regexp):
            return False
        normalized = normalize(id_number)
        checksum = MedicareNumber.checksum(id_number)
        return checksum is not None and checksum == int(normalized[8])

    @staticmethod
    def checksum(id_number: str) -> Optional[CHECK_DIGIT]:
        if not validate_regexp(id_number, MedicareNumber.METADATA.regexp):
            return None
        """algorithm: https://stackoverflow.com/questions/3589345/how-do-i-validate-an-australian-medicare-number."""
        normalized = normalize(id_number)
        # only validate first 8 digits
        number_list = [int(char) for char in list(normalized)][:8]
        total = sum([value * MedicareNumber.MAGIC_MULTIPLIER[index] for (index, value) in enumerate(number_list)])
        return total % 10



import re
from datetime import date
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..util import validate_regexp
from ..constant import Gender


def normalize(id_number: str) -> str:
    """strip out useless characters/whitespaces"""
    return re.sub(r'[.-]', '', id_number)


class ParseResult(TypedDict):
    """The parse result of Belgium NationalID"""
    yyyymmdd: date
    """birthday"""
    gender: Gender
    """gender, possible value: male and female"""
    sn: str
    """serial number"""
    checksum: int
    """checksum digits, two digits"""


class NationalRegistrationNumber:
    """
    Belgium National register number format
    https://en.wikipedia.org/wiki/Belgian_identity_card

    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'BE',
        'min_length': 11,
        'max_length': 11,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<yy>\d{2})\.?(?P<mm>\d{2})\.?(?P<dd>\d{2})-?'
                             r'(?P<sn>\d{3})\.?'
                             r'(?P<checksum>\d{2})$'),
        'alias_of': None,
        'names': ['National registration number',
                  'Belgian identity card',
                  'Identiteitskaart',
                  'Carte d’identité',
                  'Personalausweis'],
        'links': ['https://en.wikipedia.org/wiki/Belgian_identity_card'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """validate the id"""
        if not validate_regexp(id_number, NationalRegistrationNumber.METADATA.regexp):
            return False
        return NationalRegistrationNumber.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the result"""
        match_obj = NationalRegistrationNumber.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        checksum = NationalRegistrationNumber.checksum(id_number)
        if not checksum:
            return None
        else:
            yy = int(match_obj.group('yy'))
            mm = int(match_obj.group('mm'))
            dd = int(match_obj.group('dd'))
            sn = match_obj.group('sn')
            year_base = 1900 if yy > 50 else 2000
            try:
                return {
                    'yyyymmdd': date(yy + year_base, mm, dd),
                    'gender': Gender.MALE if int(sn) % 2 == 1 else Gender.FEMALE,
                    'sn': sn,
                    'checksum': int(match_obj.group('checksum'))
                }
            except ValueError:
                return None

    @staticmethod
    def checksum(id_number) -> bool:
        """
        calculated as the remainder of dividing xxxxxxxxxx by 97
        (if the remainder is 0, the check number is set to 97)
        """
        if not validate_regexp(id_number, NationalRegistrationNumber.METADATA.regexp):
            return False
        normalized = normalize(id_number)
        check_digits = normalized[-2:]
        if int(normalized[0:2]) < 50:
            # the person should be born after 2000
            check = 97 - (2000000000 + int(normalized[:-2])) % 97
        else:
            check = 97 - int(normalized[:-2]) % 97

        return int(check_digits) == check



import re
from datetime import date
from types import SimpleNamespace
from typing import Literal, Optional, TypedDict
from ..constant import Gender
from ..util import validate_regexp


class ParseResult(TypedDict):
    """The parse result of CURP"""
    name_initial_chars: str
    """initial chars of name"""
    name_consonants: str
    """consonants of name"""
    yyyymmdd: date
    """dob"""
    gender: Gender
    """gender, possible value: male, female, non-binary"""
    location: str
    """registration location"""
    sn: str
    """serial number"""
    checksum: Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    """check digit"""


class CURP:
    """
    Mexico National ID number format, CURP
    https://en.wikipedia.org/wiki/Unique_Population_Registry_Code
    http://sistemas.uaeh.edu.mx/dce/admisiones/docs/guia_CURP.pdf
    python version of https://github.com/d3249/curp
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'MX',
        'min_length': 18,
        'max_length': 18,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<initial>[A-Z]{4})'
                             r'(?P<yy>\d{2})(?P<mm>\d{2})(?P<dd>\d{2})'
                             r'(?P<gender>[HMX])'
                             r'(?P<location>[A-Z]{2})'
                             r'(?P<consonant>[A-Z]{3})'
                             r'(?P<sn>[0-9A-Z])'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['CURP',
                  'Clave Única de Registro de Población',
                  'Unique Population Registry Code',
                  'Personal ID Code Number'],
        'links': ['https://en.wikipedia.org/wiki/Unique_Population_Registry_Code',
                  'http://sistemas.uaeh.edu.mx/dce/admisiones/docs/guia_CURP.pdf'],
        'deprecated': False
    })

    ID_CHARS = '0123456789ABCDEFGHIJKLMNÑOPQRSTUVWXYZ'
    """chars index for checksum"""
    GENDER_MAP = {
        'H': Gender.MALE,
        'M': Gender.FEMALE,
        'X': Gender.NON_BINARY
    }
    """code to gender map"""

    ALLOW_LOCATIONS = ['AS', 'BC', 'BS', 'CC', 'CH', 'CL',
                       'CM', 'CS', 'DF', 'DG', 'GR', 'GT',
                       'HG', 'JC', 'MC', 'MN', 'MS', 'NE',
                       'NL', 'NT', 'OC', 'PL', 'QR', 'QT',
                       'SL', 'SP', 'SR', 'TC', 'TL', 'TS',
                       'VZ', 'YN', 'ZS']
    """possible registration location"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """validate CURP"""
        if not validate_regexp(id_number, CURP.METADATA.regexp):
            return False
        return CURP.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the result"""
        match_obj = CURP.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        location = match_obj.group('location')
        checksum = CURP.checksum(id_number)
        if not checksum:
            return None
        elif location not in CURP.ALLOW_LOCATIONS:
            return None

        yy = int(match_obj.group('yy'))
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        sn = match_obj.group('sn')
        year_base = 1900 if ord(sn) < 65 else 2000
        gender = match_obj.group('gender')
        try:
            return {
                'name_initial_chars': match_obj.group('initial'),
                'name_consonants': match_obj.group('consonant'),
                'yyyymmdd': date(yy + year_base, mm, dd),
                'gender': CURP.GENDER_MAP[gender],
                'location': match_obj.group('location'),
                'sn': sn,
                'checksum': int(match_obj.group('checksum'))
            }
        except ValueError:
            return None

    @staticmethod
    def checksum(id_number) -> bool:
        """check the checksum"""
        if not validate_regexp(id_number, CURP.METADATA.regexp):
            return False
        check = sum(CURP.ID_CHARS.index(c) * (18 - i) for i, c in enumerate(id_number[:17]))
        return int(id_number[17]) == (10 - check % 10) % 10



import re
from datetime import date
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..constant import Gender
from ..util import validate_regexp, luhn_digit


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r'[+-]', '', id_number)


class ParseResult(TypedDict):
    """parse result of PersonalIdentityNumber"""
    gender: Gender
    """gender, possible value: male, female"""
    yyyymmdd: date
    """dob"""
    checksum: str
    """checksum digit"""


class PersonalIdentityNumber:
    """
    Sweden Personal Identity number
    https://en.wikipedia.org/wiki/National_identification_number#Sweden
    https://en.wikipedia.org/wiki/Personal_identity_number_(Sweden)
    https://swedish.identityinfo.net/
    https://personnummer.dev/
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'SE',
        # length without insignificant chars
        'min_length': 10,
        'max_length': 10,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<yy>\d{2})(?P<mm>\d{2})(?P<dd>\d{2})'
                             r'(?P<sep>[+|-])'
                             r'(?!000)(?P<birth_number>\d{3})'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['Personal Identity Number',
                  'personnummer'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Sweden',
                  'https://en.wikipedia.org/wiki/Personal_identity_number_(Sweden)',
                  'https://swedish.identityinfo.net/',
                  'https://personnummer.dev/'],
        'deprecated': False

    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the SWE id number
        """
        if not validate_regexp(id_number, PersonalIdentityNumber.METADATA.regexp):
            return False
        return PersonalIdentityNumber.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        match_obj = PersonalIdentityNumber.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        checksum = match_obj.group('checksum')
        if PersonalIdentityNumber.checksum(id_number) != int(checksum):
            return None
        yy = match_obj.group('yy')
        mm = match_obj.group('mm')
        dd = match_obj.group('dd')
        birth_number = match_obj.group('birth_number')
        sep = match_obj.group('sep')
        base_year = date.today().year if sep == '-' else date.today().year - 100
        yyyy = int((base_year - ((base_year - int(yy)) % 100)) / 100) * 100 + int(yy)
        try:
            return {
                "gender": Gender.FEMALE if int(birth_number) % 2 == 0 else Gender.MALE,
                "yyyymmdd": date(yyyy, int(mm), int(dd)),
                'checksum': match_obj.group('checksum')
            }
        except ValueError:
            return None

    @staticmethod
    def checksum(id_number: str) -> Optional[int]:
        """
        algorithm: https://en.wikipedia.org/wiki/Personal_identity_number_(Sweden)#Checksum
        Multiplier start by 2
        """
        if not validate_regexp(id_number, PersonalIdentityNumber.METADATA.regexp):
            return None
        normalized = normalize(id_number)
        return luhn_digit([int(char) for char in normalized[:-1]], True)



from copy import copy
from typing import Optional, Tuple
from ..constant import Citizenship

from ..yugoslavia import ParseResult, UniqueMasterCitizenNumber as YugoslaviaJMBG

MNE_METADATA = copy(YugoslaviaJMBG.METADATA)
MNE_METADATA.iso3166_alpha2 = 'ME'


class UniqueMasterCitizenNumber(YugoslaviaJMBG):
    """
    Montenegro Unique Master Citizen Number format, JMBG
    https://en.wikipedia.org/wiki/Unique_Master_Citizen_Number
    """
    METADATA = MNE_METADATA

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the value"""
        result = YugoslaviaJMBG.parse(id_number)
        if not result:
            return None
        loc_citizenship = UniqueMasterCitizenNumber.check_location(result['location'])
        if not loc_citizenship:
            return None
        citizenship, location = loc_citizenship
        result['citizenship'] = citizenship
        return result

    @staticmethod
    def check_location(location: str) -> Optional[Tuple[Citizenship, str]]:
        result = YugoslaviaJMBG.check_location(location)
        if not result:
            return None
        """
        Since the North Macedonia is an independent country, they share the same id code base. So, the citizenship
        is 20 < location < 30
        """
        if 20 < int(location) < 30:
            return Citizenship.CITIZEN, location
        return Citizenship.RESIDENT, location


JMBG = UniqueMasterCitizenNumber
"""
Alias of UniqueMasterCitizenNumber
"""



import re
from types import SimpleNamespace
from typing import Optional
from ..util import validate_regexp


class EntityIDNumber:
    """
    The legal entity ID number in Ukraine
    https://uk.wikipedia.org/wiki/%D0%9A%D0%BE%D0%B4_%D0%84%D0%94%D0%A0%D0%9F%D0%9E%D0%A3
    https://1cinfo.com.ua/Article/Detail/Proverka_koda_po_EDRPOU/
    This is a python version of https://github.com/alazurenko/validate-edrpou
    alias: ["EDRPOU", "ЄДРПОУ"]
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'UA',
        # length without insignificant chars
        'min_length': 8,
        'max_length': 8,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^\d{8}$'),
        'alias_of': None,
        'names': ['Legal Entity ID Number',
                  'EDRPOU',
                  'ЄДРПОУ'],
        'links': ['https://uk.wikipedia.org/wiki/%D0%9A%D0%BE%D0%B4_%D0%84%D0%94%D0%A0%D0%9F%D0%9E%D0%A3',
                  'https://1cinfo.com.ua/Article/Detail/Proverka_koda_po_EDRPOU/'],
        'deprecated': False
    })

    PHASE1_MULTIPLIER = [1, 2, 3, 4, 5, 6, 7]
    PHASE2_MULTIPLIER = [7, 1, 2, 3, 4, 5, 6]

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the EDRPOU
        """
        if not validate_regexp(id_number, EntityIDNumber.METADATA.regexp):
            return False
        return EntityIDNumber.checksum(id_number) == int(id_number[7])

    @staticmethod
    def checksum(id_number: str) -> Optional[int]:
        """algorithm: https://1cinfo.com.ua/Article/Detail/Proverka_koda_po_EDRPOU/"""
        if not validate_regexp(id_number, EntityIDNumber.METADATA.regexp):
            return None
        number_list = [int(char) for char in list(id_number)]
        source_list = number_list[:7]
        if source_list[0] < 3 or source_list[0] > 6:
            multiplier = EntityIDNumber.PHASE1_MULTIPLIER
        else:
            multiplier = EntityIDNumber.PHASE2_MULTIPLIER
        modulus = sum([value * multiplier[index] for (index, value) in enumerate(source_list)]) % 11
        if modulus < 10:
            return modulus
        multiplier = [val + 2 for val in multiplier]
        # what happen when modulus is also greater than 10 in the 2nd phase?
        # implement the recursive in sequential to prevent infinite loops because I don't know what we should do
        # if the modulus is also greater than 10.
        modulus = sum([value * multiplier[index] for (index, value) in enumerate(source_list)]) % 11
        return modulus



import re
from enum import Enum
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..util import validate_regexp


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r'/\(\)', '', id_number)


class DocType(Enum):
    FIRST_GEN = 'first generation'
    MCA = 'macau civil authority'
    MPSP = 'macau public security police'
    ENTITY = 'entity'
    CI = 'commercial individual'


class ParseResult(TypedDict):
    """parse result for Identification Number"""
    doc_type: DocType
    """issuer"""
    sn: str
    """serial number"""


class NationalID:
    """
    Macau National ID number format, Permanent Resident Identity Card (BIRP),
    and Non-Permanent Resident Identity Card (BIRNP).
    https://en.wikipedia.org/wiki/National_identification_number#Macau
    https://en.wikipedia.org/wiki/Macau_Resident_Identity_Card
    https://validatetin.com/macao/
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'MO',
        'min_length': 8,
        'max_length': 8,
        'parsable': True,
        'checksum': False,
        'regexp': re.compile(r'^(?P<doc_type>[01578])'
                             r'(?P<sn>\d{6})'
                             r'\(?(?P<extra>\d)\)?$'),
        'alias_of': None,
        'names': ['National ID Number',
                  'Permanent Resident Identity Card',
                  'BIRP',
                  'Non-Permanent Resident Identity Card',
                  'BIRNP'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Macau',
                  'https://en.wikipedia.org/wiki/Macau_Resident_Identity_Card',
                  'https://validatetin.com/macao/'],
        'deprecated': False
    })

    TYPE_MAP = {'0': DocType.CI, '1': DocType.FIRST_GEN, '5': DocType.MCA, '7': DocType.MPSP, '8': DocType.ENTITY}

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the id number
        """
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        return NationalID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """pares the result"""
        match_obj = NationalID.METADATA.regexp.match(id_number)

        if not match_obj:
            return None
        doc_type = match_obj.group('doc_type')
        sn = match_obj.group('sn')
        extra = match_obj.group('extra')
        try:
            return {
                'doc_type': NationalID.TYPE_MAP[doc_type],
                'sn': sn + extra
            }
        except ValueError:
            return None



import re
from types import SimpleNamespace
from typing import Optional
from ..util import CHECK_DIGIT, validate_regexp, weighted_modulus_digit, modulus_overflow_mod10


class TaxIdentityNumber:
    """
    Greece Tax Identity Number, AFM - ΑΦΜ - Αριθμός Φορολογικού Μητρώου - Tax Registry Number
    https://en.wikipedia.org/wiki/National_identification_number#Greece
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'GR',
        'min_length': 9,
        'max_length': 9,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r'^\d{9}$'),
        'alias_of': None,
        'names': ['Tax Identity Number'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Greece'],
        'deprecated': False
    })

    MULTIPLIER = [256, 128, 64, 32, 16, 8, 4, 2]

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate with regexp
        """
        return str(TaxIdentityNumber.checksum(id_number)) == id_number[-1]

    @staticmethod
    def checksum(id_number: str) -> Optional[CHECK_DIGIT]:
        """
        ref: https://stackoverflow.com/a/4377376
        """
        if not validate_regexp(id_number, TaxIdentityNumber.METADATA.regexp):
            return None
        numbers = [int(char) for char in id_number]
        modulus = modulus_overflow_mod10(weighted_modulus_digit(numbers[0:-1], TaxIdentityNumber.MULTIPLIER, 11, True))
        return modulus



import re
from types import SimpleNamespace
from ..util import validate_regexp


class NationalID:
    """
    Argentina National ID number
    https://www.protecto.ai/argentina-national-identity-number-download-sample-data-for-testing/
    https://en.wikipedia.org/wiki/Documento_Nacional_de_Identidad_(Argentina)
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'AR',
        # length without insignificant chars
        'min_length': 8,
        'max_length': 8,
        'parsable': False,
        'checksum': False,
        'regexp': re.compile(r'^(\d{2}\.?\d{3}\.?\d{3})$'),
        'alias_of': None,
        'names': ['Documento Nacional de Identidad',
                  'DNI'],
        'links': ['https://www.protecto.ai/argentina-national-identity-number-download-sample-data-for-testing/',
                  'https://en.wikipedia.org/wiki/Documento_Nacional_de_Identidad_(Argentina)'],
        'deprecated': False

    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the ARG id number
        """
        return validate_regexp(id_number, NationalID.METADATA.regexp)



from .grc.tax_id import TaxIdentityNumber
from .grc.identity_card import IdentityCard
from .grc.old_identity_card import OldIdentityCard
from .util import alias_of

NationalID = alias_of(IdentityCard)
"""
alias of IdentityCard
"""



import os
from setuptools import setup, find_packages


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(filename):
    with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), filename), 'r') as fin:
        return fin.read()


setup(
    name='idnumbers',
    version=read('VERSION').replace('\n', '').replace('\r', ''),
    author='Identique',
    author_email='microdataxyz@outlook.com',
    description='id numbers verification toolkits',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    license='MIT',
    url='https://github.com/Identique/idnumbers',
    project_urls={
        'Source': 'https://github.com/Identique/idnumbers',
        'Tracker': 'https://github.com/Identique/idnumbers/issues',
    },
    packages=find_packages(exclude=['*tests*']),
    data_files=[('version', ['VERSION'])],
    python_requires='>=3.9',
    install_requires=[],
    setup_requires=[],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'Typing :: Typed'
    ], )



import re
from datetime import date
from types import SimpleNamespace
from typing import TypedDict, Optional

from ..util import validate_regexp, CHECK_DIGIT, weighted_modulus_digit
from ..constant import Gender


class ParseResult(TypedDict):
    """Parse result of UniformCivilNumber"""
    yyyymmdd: date
    """birthday"""
    checksum: CHECK_DIGIT
    """check digits"""
    gender: Gender
    """gender: male/female"""


class UniformCivilNumber:
    """
    Bulgaria Uniform civil number
    https://en.wikipedia.org/wiki/National_identification_number#Bulgaria
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'BG',
        # length without insignificant chars
        'min_length': 10,
        'max_length': 10,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<yy>\d{2})'
                             r'(?P<mm>\d{2})'
                             r'(?P<dd>\d{2})'
                             r'\d{2}'
                             r'(?P<gender>\d)'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['Uniform civil number',
                  'Единен граждански номер',
                  'Edinen grazhdanski nomer',
                  'ЕГН',
                  'EGN'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Bulgaria'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the BGR id number
        """
        if not validate_regexp(id_number, UniformCivilNumber.METADATA.regexp):
            return False
        return UniformCivilNumber.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """
        Parse the result
        """
        match_obj = UniformCivilNumber.METADATA.regexp.match(id_number)
        checksum = int(match_obj.group("checksum"))
        yy = int(match_obj.group("yy"))
        mm = int(match_obj.group("mm"))
        dd = int(match_obj.group("dd"))
        if UniformCivilNumber.checksum(id_number) != checksum:
            return None
        if mm > 40:
            mm -= 40
            yyyy = yy + 2000
        elif mm > 20:
            mm -= 20
            yyyy = yy + 1800
        else:
            yyyy = yy + 1900
        try:
            return {
                'yyyymmdd': date(yyyy, mm, dd),
                "checksum": int(checksum),
                'gender': Gender.MALE if int(match_obj.group("gender")) % 2 == 0 else Gender.FEMALE
            }
        except ValueError:
            return None

    MULTIPLIER = [2, 4, 8, 5, 10, 9, 7, 3, 6]

    @staticmethod
    def checksum(id_number: str) -> CHECK_DIGIT:
        """
        Get the checksum digit
        https://en.wikipedia.org/wiki/Unique_citizenship_number
        """
        digits_numbers = [int(i) for i in id_number[:-1]]
        return weighted_modulus_digit(digits_numbers, UniformCivilNumber.MULTIPLIER, 11, True)



import re
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..util import CHECK_DIGIT


class ParseResult(TypedDict):
    yymm: str
    """birthday of this ID"""
    sn: str
    """serial number"""
    checksum: CHECK_DIGIT
    """checksum"""


class PersonalNumber:
    """
    Bahrain Personal Number, Identification Card Number, Arabic:  الرقم الشخصي
    https://en.wikipedia.org/wiki/National_identification_number#Bahrain
    * According to the doc, we can know it has checksum algorithm. But we cannot find it.
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'BH',
        'min_length': 9,
        'max_length': 9,
        'parsable': True,
        'checksum': False,
        'regexp': re.compile(r'^(?P<yymm>\d{2}(?:0[1-9]|1[012]))'
                             r'(?P<sn>\d{4})'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['Personal number',
                  'Identification card number',
                  'بطاقة الهوية',
                  'الرقم الشخصي',
                  'Central population registration number',
                  'CPR',
                  'الرقم السكاني'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Bahrain'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the civil number
        """
        if not id_number:
            return False

        if not isinstance(id_number, str):
            id_number = repr(id_number)
        return PersonalNumber.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """
        parse the id number
        """
        match_obj = PersonalNumber.METADATA.regexp.match(id_number)
        if not match_obj:
            return None

        # TODO: find and implement checksum
        return {
            'yymm': match_obj.group('yymm'),
            'sn': match_obj.group('sn'),
            'checksum': int(match_obj.group('checksum'))
        }



import re
from enum import Enum
from types import SimpleNamespace
from typing import TypedDict, Optional

from ..util import validate_regexp


class ResidentialType(Enum):
    RURAL = 1
    MUNICIPALITY = 2
    CITY = 3
    OTHERS = 4
    CANTONMENT = 5
    CITY_CORPORATION = 9


class OldParseResult(TypedDict):
    distinct: str
    """distinct code"""
    residential_type: ResidentialType
    """RMO/residential type"""
    policy_station_no: str
    """upazilla/ thana/police station number"""
    union_code: str
    """your union code/ward number"""
    sn: str
    """serial no"""


class OldNationalID:
    """
    Bangladesh National ID number, জাতীয় পরিচয়পত্র, NID, BD
    https://en.wikipedia.org/wiki/National_identity_card_(Bangladesh)
    http://nationalidcardbangladesh.blogspot.com/2016/04/voter-id-national-id-card-number.html
    https://www.facebook.com/428195627559147/photos/a.428251897553520/428251617553548/?type=3
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'BD',
        # length without insignificant chars
        'min_length': 13,
        'max_length': 13,
        'parsable': True,
        'checksum': False,
        'regexp': re.compile(r'^(?P<distinct>\d{2})'
                             r'(?P<rmo>\d)'
                             r'(?P<police>\d{2})'
                             r'(?P<union>\d{2})'
                             r'(?P<sn>\d{6})$'),
        'alias_of': None,
        'names': ['Bangladesh national ID number',
                  'জাতীয় পরিচয়পত্র',
                  'NID',
                  'BD'],
        'links': [
            'https://en.wikipedia.org/wiki/National_identity_card_(Bangladesh)',
            'http://nationalidcardbangladesh.blogspot.com/2016/04/voter-id-national-id-card-number.html',
            'https://www.facebook.com/428195627559147/photos/a.428251897553520/428251617553548/?type=3'],
        'deprecated': True
    })

    RMO_MAP = {
        '1': ResidentialType.RURAL,
        '2': ResidentialType.MUNICIPALITY,
        '3': ResidentialType.CITY,
        '4': ResidentialType.OTHERS,
        '5': ResidentialType.CANTONMENT,
        '9': ResidentialType.CITY_CORPORATION
    }

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate
        """
        if not validate_regexp(id_number, OldNationalID.METADATA.regexp):
            return False
        return OldNationalID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[OldParseResult]:
        """
        Parse the result
        """
        match_obj = OldNationalID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None

        if match_obj.group('rmo') not in OldNationalID.RMO_MAP:
            return None

        try:
            return {
                'distinct': match_obj.group('distinct'),
                'residential_type': OldNationalID.RMO_MAP[match_obj.group('rmo')],
                'policy_station_no': match_obj.group('police'),
                'union_code': match_obj.group('union'),
                'sn': match_obj.group('sn')
            }
        except ValueError:
            return None



import re
from types import SimpleNamespace

from ..util import validate_regexp


class NationalID:
    """
    Hong Kong national ID number format
    https://en.wikipedia.org/wiki/National_identification_number#Hong_Kong
    https://pinkylam.me/playground/hkid/
    https://github.com/hsyuen720/hkid-tools/blob/main/app/utils/validate.ts
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'HK',
        # length without insignificant chars
        'min_length': 8,
        'max_length': 9,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r"^[A-Z]{1,2}[0-9]{6}[0-9A]$"),
        'alias_of': None,
        'names': ['National ID Number',
                  '香港身份證'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Hong_Kong',
                  'https://pinkylam.me/playground/hkid/'],
        'deprecated': False

    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate HKG id number
        """
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        return NationalID.checksum(id_number) == id_number[-1]

    @staticmethod
    def checksum(id_number: str) -> str:
        """
        Calculate HKG national id checksum digit
        """
        arr = list(id_number[:-1])
        multiplier = [len(arr) + 1 - index for (index, _) in enumerate(arr)]
        total = 0 if len(arr) % 2 == 0 else 36 * 9
        total += sum([NationalID.get_number(arr[idx]) * mul for (idx, mul) in enumerate(multiplier)])
        rem = total % 11
        return "A" if rem == 1 else '0' if rem == 0 else str(11 - rem)

    @staticmethod
    def get_number(digit: str) -> int:
        """Convert letter to number"""
        if re.match('[A-Z]', digit):
            return int(ord(digit)) - 55
        else:
            return int(digit)



import re
from datetime import date
from math import floor
from types import SimpleNamespace
from typing import Optional, Tuple, TypedDict
from ..constant import Gender
from ..util import CHECK_DIGIT, validate_regexp


class ParseResult(TypedDict):
    yyyymmdd: date
    """birthday of this ID"""
    gender: Gender
    """only male or female"""
    sn: str
    """serial number"""
    checksum: CHECK_DIGIT
    """checksum digits"""


class PersonalCode:
    """
    Lithuania personal code, asmens kodas
    https://en.wikipedia.org/wiki/National_identification_number#Lithuania
    https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/Lithuania-TIN.pdf
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'LT',
        'min_length': 11,
        'max_length': 11,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<g>\d)'
                             r'(?P<yy>\d{2})(?P<mm>\d{2})(?P<dd>\d{2})'
                             r'(?P<sn>\d{3})'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['Personal Code',
                  'asmens kodas'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Lithuania',
                  'https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/'
                  'tax-identification-numbers/Lithuania-TIN.pdf'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the Italy fiscal code
        """
        if not id_number:
            return False

        if not isinstance(id_number, str):
            id_number = repr(id_number)
        return PersonalCode.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """
        parse the id number
        """
        match_obj = PersonalCode.METADATA.regexp.match(id_number)
        if not match_obj:
            return None

        checksum = PersonalCode.checksum(id_number)
        if checksum is None or str(checksum) != match_obj.group('checksum'):
            return None
        year_base, gender = PersonalCode.extract_year_base_gender(int(match_obj.group('g')))
        yy = int(match_obj.group('yy'))
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        try:
            return {
                'yyyymmdd': date(year_base + yy, mm, dd),
                'gender': gender,
                'sn': match_obj.group('sn'),
                'checksum': checksum
            }
        except ValueError:
            return None

    @staticmethod
    def checksum(id_number) -> Optional[CHECK_DIGIT]:
        """
        algorithm https://en.wikipedia.org/wiki/National_identification_number#Lithuania
        """
        if not validate_regexp(id_number, PersonalCode.METADATA.regexp):
            return None
        b = 1
        c = 3
        d = 0
        e = 0
        numbers = [int(char) for char in id_number]
        for number in numbers[:-1]:
            d += number * b
            e += number * c
            b = b + 1 if b < 9 else 1
            c = c + 1 if c < 9 else 1
        d %= 11
        e %= 11
        if d < 10:
            return d
        elif e < 10:
            return e
        else:
            return 0

    @staticmethod
    def extract_year_base_gender(g: CHECK_DIGIT) -> Optional[Tuple[int, Gender]]:
        """
        algorithm: G = floor(year / 100) * 2 - 34 - gender while gender = {female: 0, male: 1}
        if the value is odd -> male, if the value is even -> female
        """
        gender = Gender.FEMALE if g % 2 == 0 else Gender.MALE
        # remove the affect of gender, -1 if it is male otherwise it should be the original value
        year_g = (g // 2) * 2
        year_base = floor((year_g + 34) / 2) * 100
        return year_base, gender



import re
from types import SimpleNamespace
from idnumbers.nationalid.util import validate_regexp


class NationalID:
    """
    Nigeria national ID number format
    https://en.wikipedia.org/wiki/National_identification_number#Nigeria
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'NG',
        # length without insignificant chars
        'min_length': 11,
        'max_length': 11,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': False,
        # regular expression to validate the id
        'regexp': re.compile(r'^\d{11}$'),
        'alias_of': None,
        'names': ['National Identification Number',
                  'NIN'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Nigeria'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the NGA id number
        """
        return validate_regexp(id_number, NationalID.METADATA.regexp)



import re
from datetime import date
from types import SimpleNamespace
from typing import Optional, Tuple, TypedDict
from ..constant import Gender
from ..util import CHECK_DIGIT, validate_regexp, weighted_modulus_digit


class ParseResult(TypedDict):
    """The parse result of Resident ID"""
    yyyymmdd: date
    """birthday"""
    sn: str
    """serial number"""
    gender: Gender
    """gender, possible value: male, female"""
    checksum: CHECK_DIGIT
    """checksum digit, numbers or 'X'"""


class PersonalID:
    """
    Estonia Personal ID number format, isikukood
    https://en.wikipedia.org/wiki/National_identification_number#Estonia
    https://et.wikipedia.org/wiki/Isikukood
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'EE',
        'min_length': 11,
        'max_length': 11,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<gender_century>\d)'
                             r'(?P<yy>\d{2})(?P<mm>\d{2})(?P<dd>\d{2})'
                             r'(?P<sn>\d{3})'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['Personal ID Number',
                  'isikukood'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Estonia',
                  'https://et.wikipedia.org/wiki/Isikukood'],
        'deprecated': False
    })

    WEIGHTS1 = [1, 2, 3, 4, 5, 6, 7, 8, 9, 1]
    WEIGHTS2 = [3, 4, 5, 6, 7, 8, 9, 1, 2, 3]

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the personal id number
        """
        if not id_number:
            return False

        if not isinstance(id_number, str):
            id_number = repr(id_number)
        return PersonalID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the data"""
        match_obj = PersonalID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        checksum_str = match_obj.group('checksum')
        checksum = PersonalID.checksum(id_number)
        if checksum is None or str(checksum) != checksum_str:
            return None
        gender_year_base = PersonalID.get_gender_year_base(int(match_obj.group('gender_century')))
        if gender_year_base is None:
            return None
        gender, year_base = gender_year_base
        try:
            sn = match_obj.group('sn')
            return {
                'yyyymmdd': date(int(match_obj.group('yy')) + year_base,
                                 int(match_obj.group('mm')),
                                 int(match_obj.group('dd'))),
                'sn': sn,
                'gender': gender,
                'checksum': checksum
            }
        except ValueError:
            return None

    @staticmethod
    def get_gender_year_base(gender_century: int) -> Optional[Tuple[Gender, int]]:
        gender = Gender.MALE if gender_century % 2 == 1 else Gender.FEMALE
        if 0 < gender_century <= 2:
            year_base = 1800
        elif 2 < gender_century <= 4:
            year_base = 1900
        elif 4 < gender_century <= 6:
            year_base = 2000
        elif 6 < gender_century <= 8:
            year_base = 2100
        else:
            return None
        return gender, year_base

    @staticmethod
    def checksum(id_number) -> Optional[CHECK_DIGIT]:
        """algorithm: https://et.wikipedia.org/wiki/Isikukood#Kontrollnumber"""
        if not validate_regexp(id_number, PersonalID.METADATA.regexp):
            return None
        numbers = [int(char) for char in id_number]
        checksum = weighted_modulus_digit(numbers[0:-1], PersonalID.WEIGHTS1, 11, True)
        if checksum == 10:
            # use 2 phase weights when it is 10
            checksum = weighted_modulus_digit(numbers[0:-1], PersonalID.WEIGHTS2, 11, True)
            if checksum == 10:
                # reset to 0 if it is 10 at the 2nd phase
                checksum = 0
        return checksum



from enum import Enum


class Gender(Enum):
    """general gender enum"""
    MALE = 'male'
    FEMALE = 'female'
    NON_BINARY = 'non_binary'
    """not all countries have this gender type."""


class Citizenship(Enum):
    CITIZEN = 'citizen'
    RESIDENT = 'resident'
    FOREIGN = 'foreign'
    """foreign may be naturalized citizen or a resident"""



import re
from types import SimpleNamespace
from ..util import validate_regexp


class PersonalNumber:
    """
    Georgia personal number format
    https://en.wikipedia.org/wiki/National_identification_number#Georgia
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'GE',
        'min_length': 9,
        'max_length': 9,
        'parsable': False,
        'checksum': False,
        'regexp': re.compile(r'^\d{9}$'),
        'alias_of': None,
        'names': ['personal number'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Georgia'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate personal number
        """
        return validate_regexp(id_number, PersonalNumber.METADATA.regexp)



import re
from datetime import date
from types import SimpleNamespace
from typing import Literal, Optional, TypedDict
from ..constant import Gender


class ParseResult(TypedDict):
    """Parse result of id number"""
    yyyymmdd: date
    """year of birth"""
    sn: str
    """serial number"""
    gender: Gender
    """gender: male or female"""
    checksum: Literal['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K',
                      'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'Y', 'U',
                      'V', 'W']  # cannot use CHECK_ALPHA because it ends with W
    """check digits"""


class IdentityNumber:
    """
    Albania Identity Number, Numri i Identitetit (NID),  Numri i Identitetit të Shtetasit (NISH), NIPT
    https://en.wikipedia.org/wiki/National_identification_number#Albania
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'AL',
        'min_length': 10,
        'max_length': 10,
        'parsable': True,
        'checksum': False,  # There is a checksum algorithm. But we cannot find it.
        'regexp': re.compile(r'^(?P<yy>[0-9A-T]\d)(?P<mm>\d{2})(?P<dd>\d{2})'
                             r'(?P<sn>\d{3})[ -]?'
                             r'(?P<checksum>[A-W])$'),
        'alias_of': None,
        'names': ['Albania Identity Number',
                  'Numri i Identitetit',
                  'NID',
                  'Numri i Identitetit të Shtetasit'
                  'NISH',
                  'NIPT'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Albania'],
        'deprecated': False
    })

    BASE_YEAR_MAP = '0123456789ABCDEFGHIJKLMNOPQRST'

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the id number
        """
        if not id_number:
            return False

        if not isinstance(id_number, str):
            id_number = repr(id_number)
        return IdentityNumber.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the result"""
        match_obj = IdentityNumber.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        yyyy = IdentityNumber.get_year(match_obj.group('yy'))
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        try:
            return {
                'yyyymmdd': date(yyyy, mm if mm < 50 else mm - 50, dd),
                'gender': Gender.MALE if mm < 50 else Gender.FEMALE,
                'sn': match_obj.group('sn'),
                'checksum': match_obj.group('checksum')
            }
        except ValueError:
            return None

    @staticmethod
    def get_year(yy: str) -> int:
        year_base = 1800 + IdentityNumber.BASE_YEAR_MAP.index(yy[0]) * 10
        return year_base + int(yy[1])



import re
from datetime import date
from types import SimpleNamespace
from typing import Optional, Tuple, TypedDict, cast
from ..constant import Gender
from ..util import CHECK_ALPHA, validate_regexp


class ParseResult(TypedDict):
    """parse result of FiscalCode"""
    surname_consonants: str
    """The 3 consonants chars of surname"""
    firstname_consonants: str
    """The 3 consonants chars of firstname"""
    yyyymmdd: date
    """birthday of this ID"""
    gender: Gender
    """only male or female"""
    area_code: str
    """registry code check the page https://en.wikipedia.org/wiki/Italian_fiscal_code"""
    checksum: CHECK_ALPHA
    """alphabet checksum"""


class FiscalCode:
    """
    Italy fiscal code
    https://en.wikipedia.org/wiki/Italian_fiscal_code
    https://en.wikipedia.org/wiki/National_identification_number#Italy
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'IT',
        'min_length': 16,
        'max_length': 16,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<surname>[A-Z]{3})'
                             r'(?P<firstname>[A-Z]{3})'
                             r'(?P<yy>[0-9A-Z]{2})'
                             r'(?P<m>[A-EHLMPR-T])'
                             r'(?P<dd>[0-9A-Z]{2})'
                             r'(?P<area_code>[A-Z][0-9A-Z]{3})'
                             r'(?P<checksum>[A-Z])$'),
        'alias_of': None,
        'names': ['Fiscal Code',
                  'Codice fiscale'],
        'links': ['https://en.wikipedia.org/wiki/Italian_fiscal_code',
                  'https://en.wikipedia.org/wiki/National_identification_number#Italy'],
        'deprecated': False
    })

    MONTH_MAP = {'A': 1, 'B': 2, 'C': 3, 'D': 4,
                 'E': 5, 'H': 6, 'L': 7, 'M': 8,
                 'P': 9, 'R': 10, 'S': 11, 'T': 12}
    """char to month map"""

    MAGIC_ODD_CHAR_MAP = {'0': 1, '9': 21, 'I': 19, 'R': 8,
                          '1': 0, 'A': 1, 'J': 21, 'S': 12,
                          '2': 5, 'B': 0, 'K': 2, 'T': 14,
                          '3': 7, 'C': 5, 'L': 4, 'U': 16,
                          '4': 9, 'D': 7, 'M': 18, 'V': 10,
                          '5': 13, 'E': 9, 'N': 20, 'W': 22,
                          '6': 15, 'F': 13, 'O': 11, 'X': 25,
                          '7': 17, 'G': 15, 'P': 3, 'Y': 24,
                          '8': 19, 'H': 17, 'Q': 6, 'Z': 23}
    """magic numbers for odd chars"""

    MAGIC_EVEN_CHAR_MAP = {'0': 0, '9': 9, 'I': 8, 'R': 17,
                           '1': 1, 'A': 0, 'J': 9, 'S': 18,
                           '2': 2, 'B': 1, 'K': 10, 'T': 19,
                           '3': 3, 'C': 2, 'L': 11, 'U': 20,
                           '4': 4, 'D': 3, 'M': 12, 'V': 21,
                           '5': 5, 'E': 4, 'N': 13, 'W': 22,
                           '6': 6, 'F': 5, 'O': 14, 'X': 23,
                           '7': 7, 'G': 6, 'P': 15, 'Y': 24,
                           '8': 8, 'H': 7, 'Q': 16, 'Z': 25}
    """magic numbers for even chars"""

    NUMERIC_REPLACEMENT = {'L': '0', 'Q': '4', 'U': '8',
                           'M': '1', 'R': '5', 'V': '9',
                           'N': '2', 'S': '6',
                           'P': '3', 'T': '7'}
    """reverse numbers replacement for conflict IDs"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the Italy fiscal code
        """
        if not id_number:
            return False

        if not isinstance(id_number, str):
            id_number = repr(id_number)
        return FiscalCode.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """
        parse the id number
        """
        match_obj = FiscalCode.METADATA.regexp.match(id_number)
        if not match_obj:
            return None

        # sterilization from right to left. we need to do the area_code first
        area_code = match_obj.group('area_code')
        sterilized_ac_nums = FiscalCode.sterilize_numbers(area_code[1:])
        if not sterilized_ac_nums:
            return None

        # let the extract_birthday to sterilize others
        dob_gender = FiscalCode.extract_birthday(match_obj.group('yy'),
                                                 match_obj.group('m'),
                                                 match_obj.group('dd'))
        if dob_gender is None:
            return None

        checksum = FiscalCode.checksum(id_number)
        if checksum is None or str(checksum) != match_obj.group('checksum'):
            return None

        return {
            'surname_consonants': match_obj.group('surname'),
            'firstname_consonants': match_obj.group('firstname'),
            'area_code': area_code[0] + sterilized_ac_nums,
            'yyyymmdd': dob_gender[0],
            'gender': dob_gender[1],
            'checksum': checksum
        }

    @staticmethod
    def checksum(id_number) -> Optional[CHECK_ALPHA]:
        """
        build the checksum after the sterilization
        """
        if not validate_regexp(id_number, FiscalCode.METADATA.regexp):
            return None
        odd_total = 0
        even_total = 0
        alphanum = id_number[:-1]
        for index, char in enumerate(alphanum):
            if (index + 1) % 2 == 1:
                odd_total += FiscalCode.MAGIC_ODD_CHAR_MAP[char]
            else:
                even_total += FiscalCode.MAGIC_EVEN_CHAR_MAP[char]
        modulus = (odd_total + even_total) % 26
        return cast(CHECK_ALPHA, chr(65 + modulus))

    @staticmethod
    def extract_birthday(yy_str: str, m: str, dd_str: str) -> Optional[Tuple[date, Gender]]:
        """sterilize the numbers and convert the str to DoB and gender"""
        if m not in FiscalCode.MONTH_MAP:
            return None
        sterilized_dd = FiscalCode.sterilize_numbers(dd_str)
        if not sterilized_dd:
            return None
        sterilized_yy = FiscalCode.sterilize_numbers(yy_str)
        if not sterilized_yy:
            return None
        yy = int(sterilized_yy)
        dd = int(sterilized_dd)
        year_base = 2000 if yy < 50 else 1900
        mm = FiscalCode.MONTH_MAP[m]
        day = dd if dd < 40 else dd - 40
        gender = Gender.MALE if dd < 40 else Gender.FEMALE
        try:
            return date(year_base + yy, mm, day), gender
        except ValueError:
            return None

    @staticmethod
    def sterilize_numbers(source: str) -> Optional[str]:
        """
        When the id is conflict with others, it replaces numbers to char. We need to reverse them.
        """
        result = ''
        for char in list(source[::-1]):
            if ord(char) > 64:
                if char not in FiscalCode.NUMERIC_REPLACEMENT:
                    return None
                result += char
            else:
                result += char

        return result[::-1]



import re
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..util import validate_regexp


class ParseResult(TypedDict):
    """Parse result of Zimbabwe national identity code"""
    register_office_code: str
    """Register office code"""
    checksum: str
    """Checksum code"""
    district_code: str
    """Code of the district"""


class NationalID:
    """
    Zimbabwe National ID number format
    https://en.wikipedia.org/wiki/National_identification_number#Zimbabwe
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'ZW',
        'min_length': 11,
        'max_length': 12,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<register_office_code>\d{2})'
                             r'(?P<national_num>(\d{6}|\d{7}))'
                             r'(?P<checksum>[A-Z])'
                             r'(?P<district_code>\d{2}$)'),
        'alias_of': None,
        'names': ['National ID Number'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Zimbabwe'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the ZWE id number
        """
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        elif NationalID.parse(id_number) is None:
            return False
        return NationalID.checksum(id_number)

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the ZWE national id"""
        match_obj = NationalID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        register_office_code = match_obj.group('register_office_code')
        checksum = match_obj.group('checksum')
        district_code = match_obj.group('district_code')
        if not NationalID.checksum(id_number):
            return None
        elif not NationalID.check_district_code(register_office_code):
            return None
        # 00 is valid district code for foreigner
        elif not NationalID.check_district_code(district_code) and district_code != '00':
            return None
        else:
            return {
                'register_office_code': register_office_code,
                'checksum': checksum,
                'district_code': district_code
            }

    @staticmethod
    def checksum(id_number) -> bool:
        """Validate checksum"""
        match_obj = NationalID.METADATA.regexp.match(id_number)
        register_office_code = match_obj.group('register_office_code')
        national_num = match_obj.group('national_num')
        checksum_code = match_obj.group('checksum')
        return NationalID.get_checksum(register_office_code, national_num) == checksum_code

    @staticmethod
    def get_checksum(register_office_code: str, national_num: str) -> str:
        """
        Implement the checksum rule by
        https://www.slideshare.net/povonews/zimbabwe-2018-biometric-voters-roll-analysis-pachedu
        page 56 Appendix 2
        """
        remainder = sum(int(d) for d in (register_office_code + national_num)) % 23
        checksum_list = ['Z', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'J', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T',
                         'V', 'W', 'X', 'Y']
        return checksum_list[remainder]

    @staticmethod
    def check_district_code(code: str) -> bool:
        """Map the district code"""
        valid_district_codes = ['02', '03', '04', '05', '06', '07', '08', '10', '11', '12', '13', '14', '15', '18',
                                '19', '21', '22', '23', '24', '25', '26', '27', '28', '29', '32', '34', '35', '37',
                                '38', '39', '41', '42', '43', '44', '45', '46', '47', '48', '49', '50', '53', '54',
                                '56', '58', '59', '61', '63', '66', '67', '68', '70', '71', '73', '75', '77', '79',
                                '80', '83', '84', '85', '86']
        return code in valid_district_codes



import re
from types import SimpleNamespace

from ..util import validate_regexp


class SocialInsuranceNumber:
    """
    Canada social insurance number format
    https://en.wikipedia.org/wiki/National_identification_number#Canada
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'CA',
        # length without insignificant chars
        'min_length': 9,
        'max_length': 9,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^\d{9}$'),
        'alias_of': None,
        'names': ['Social Insurance Number',
                  'SIN'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Canada']
    })

    MULTIPLIER = [1, 2, 1, 2, 1, 2, 1, 2, 1]

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the CAN id number
        """
        if not validate_regexp(id_number, SocialInsuranceNumber.METADATA.regexp):
            return False
        return SocialInsuranceNumber.checksum(id_number)

    @staticmethod
    def checksum(id_number: str) -> bool:
        """
        Validate social insurance number checksum digits
        http://www.straightlineinternational.com/docs/vaildating_canadian_sin.pdf
        """

        number_list = [int(char) for char in list(id_number)]
        multiplied_list = [value * SocialInsuranceNumber.MULTIPLIER[index] for (index, value) in enumerate(number_list)]
        return sum([sum(divmod(num, 10)) for num in multiplied_list]) % 10 == 0



import re
from types import SimpleNamespace
from ..util import validate_regexp


class DNI:
    """
    Spain National ID number
    Documento Nacional de Identidad (DNI)
    https://en.wikipedia.org/wiki/National_identification_number#Spain
    https://es.wikipedia.org/wiki/C%C3%B3digo_de_identificaci%C3%B3n_fiscal
    https://gist.github.com/afgomez/5691823
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'ES',
        # length without insignificant chars
        'min_length': 9,
        'max_length': 9,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r'^(\d{8})([A-Z])$'),
        'alias_of': None,
        'names': ['Documento Nacional de Identidad',
                  'DNI'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Spain',
                  'https://es.wikipedia.org/wiki/C%C3%B3digo_de_identificaci%C3%B3n_fiscal'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the Spain national id number
        """
        if not validate_regexp(id_number, DNI.METADATA.regexp):
            return False
        return DNI.checksum(id_number)

    MAGIC_LETTERS = 'TRWAGMYFPDXBNJZSQVHLCKE'

    @staticmethod
    def checksum(id_number: str) -> bool:
        """algorithm: https://en.wikipedia.org/wiki/Documento_Nacional_de_Identidad_(Spain)#Number"""
        if not validate_regexp(id_number, DNI.METADATA.regexp):
            return False
        idx = int(id_number[:-1]) % 23
        return DNI.MAGIC_LETTERS[idx] == id_number[-1]



import re
from types import SimpleNamespace
from ..util import validate_regexp, weighted_modulus_digit


class TaxIDNumber:
    """
    Portugal Tax ID number
    Número de identificação fiscal or NIF
    https://en.wikipedia.org/wiki/National_identification_number#Portugal
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'PT',
        # length without insignificant chars
        'min_length': 9,
        'max_length': 9,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r'^([12356][0-9]|45|7[012]|9[0189])\d{7}$'),
        'alias_of': None,
        'names': ['Tax ID Number',
                  'Número de identificação fiscal'
                  'NIF'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Portugal'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the Tax ID number
        https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/Portugal-TIN.pdf
        """
        return TaxIDNumber.checksum(id_number)

    @staticmethod
    def checksum(id_number: str) -> bool:
        """check the checksum"""
        if not validate_regexp(id_number, TaxIDNumber.METADATA.regexp):
            return False
        multipliers = [9, 8, 7, 6, 5, 4, 3, 2]
        mod = weighted_modulus_digit([int(i) for i in id_number[:-1]], multipliers, 11, True)
        calculated_checksum = 0 if mod == 0 or mod == 1 else (11 - mod)
        return calculated_checksum == int(id_number[-1])



import re
from datetime import date
from types import SimpleNamespace
from typing import Optional, TypedDict, Tuple
from .util import CHECK_DIGIT, weighted_modulus_digit, validate_regexp
from .constant import Citizenship, Gender


class ParseResult(TypedDict):
    """
    parse result of JMBG
    """
    yyyymmdd: date
    """date of birth"""
    location: str
    """birth location"""
    citizenship: Citizenship
    """citizenship of Slovenia"""
    gender: Gender
    """gender, male or female"""
    sn: str
    """serial"""
    checksum: CHECK_DIGIT
    """checksum digit"""


class UniqueMasterCitizenNumber:
    """
    Yugoslavia JMBG which is shared among all independent countries from Yugoslavia.
    https://en.wikipedia.org/wiki/Unique_Master_Citizen_Number
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': None,
        'min_length': 13,
        'max_length': 13,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<dd>\d{2})'
                             r'(?P<mm>\d{2})'
                             r'(?P<yyy>\d{3})'
                             r'(?P<location>\d{2})'
                             r'(?P<sn>\d{3})'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['Unique  master citizen number',
                  'JMBG',
                  'Jedinstveni matični broj građana',
                  'Јединствени матични број грађана',
                  'ЈМБГ',
                  'Единствен матичен број на граѓанинот',
                  'ЕМБГ',
                  'Enotna matična številka občana,',
                  'EMŠO'],
        'links': ['https://en.wikipedia.org/wiki/Unique_Master_Citizen_Number'],
        'deprecated': False
    })

    MAGIC_MULTIPLIER = [7, 6, 5, 4, 3, 2]
    """multiplier for checksum"""

    LOC_BLACK_LIST = ['20', '40', '51', '52', '53', '54', '55', '56', '57', '58', '59', '90', '97', '98', '99']

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the JMBG id number
        """
        if not validate_regexp(id_number, UniqueMasterCitizenNumber.METADATA.regexp):
            return False
        return UniqueMasterCitizenNumber.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the value"""
        match_obj = UniqueMasterCitizenNumber.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        checksum = UniqueMasterCitizenNumber.checksum(id_number)
        if not checksum:
            return None
        loc_citizenship = UniqueMasterCitizenNumber.check_location(match_obj.group('location'))
        if not loc_citizenship:
            return None
        citizenship, location = loc_citizenship
        yyy = int(match_obj.group('yyy'))
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        year_base = 2000 if yyy < 800 else 1000
        sn = match_obj.group('sn')
        try:
            return {
                'yyyymmdd': date(year_base + yyy, mm, dd),
                'location': location,
                'citizenship': citizenship,
                'gender': Gender.MALE if int(sn) < 500 else Gender.FEMALE,
                'sn': sn,
                'checksum': int(match_obj.group('checksum'))
            }
        except ValueError:
            return None

    @staticmethod
    def checksum(id_number) -> bool:
        """
        algorithm:
        https://en.wikipedia.org/wiki/Unique_Master_Citizen_Number#Checksum_calculation
        """
        if not validate_regexp(id_number, UniqueMasterCitizenNumber.METADATA.regexp):
            return False
        numbers = [int(char) for char in id_number]
        # fold the first 12 digits
        folded = []
        for idx in range(6):
            folded.append(numbers[idx] + numbers[idx + 6])
        # it uses modulus 10 algorithm with magic numbers
        modulus = weighted_modulus_digit(folded, UniqueMasterCitizenNumber.MAGIC_MULTIPLIER, 11)
        if modulus > 9:
            modulus = 0
        return modulus == numbers[-1]

    @staticmethod
    def check_location(location: str) -> Optional[Tuple[Citizenship, str]]:
        """
        Check location with the list from here
        https://en.wikipedia.org/wiki/Unique_Master_Citizen_Number#Composition
        """
        if location in UniqueMasterCitizenNumber.LOC_BLACK_LIST:
            return None
        return Citizenship.CITIZEN, location



from .smr.social_security import SocialSecurityNumber
from .smr.tax_registration import TaxRegistrationNumber
from .util import alias_of

NationalID = alias_of(SocialSecurityNumber)
"""
alias of SocialSecurityNumber
"""
